;;; -*- lexical-binding:t -*-

(defun window-tree-pixel-sizes (window &optional next)
  "Return pixel sizes of all windows rooted at WINDOW.
The return value is a list where each window is represented either by a
triple whose first element is either t for an internal window that is a
horizontal combination, nil for an internal window that is a vertical
combination, or the window itself for a valid window.  The second element
is a cons of the pixel height and pixel width of the window.  The third
element is specified for internal windows only and recursively lists
that window's child windows using the same triple structure."
  (let (list)
    (while window
      (setq list
	    (cons
	     (cond
	      ((window-top-child window)
	       (cons (cons t window) (cons (cons (window-normal-size window nil)
				   (window-normal-size window t))
			     (window-tree-pixel-sizes
			      (window-top-child window) t))))
	      ((window-left-child window)
	       (cons (cons nil window) (cons (cons (window-normal-size window nil)
				     (window-normal-size window t))
			       (window-tree-pixel-sizes
				(window-left-child window) t))))
	      (t (list window (cons (window-normal-size window nil)
				    (window-normal-size window t)))))
	     list))
      (setq window (when next (window-next-sibling window))))
    (nreverse list)))

(defun rotate-windows-anticlockwise (&optional subtree frame-or-window)
  "Rotate windows of FRAME-OR-WINDOW anticlockwise by 90 degrees.
FRAME-OR-WINDOW must be a valid frame or window and defaults to the
selected frame.  If FRAME-OR-WINDOW is a frame, rotate the main window
of the frame, otherwise rotate FRAME-OR-WINDOW.  See
`rotate-windows-clockwise' for how to rotate windows in the opposite
direction.  If SUBTREE is non-nil, the function will act as if
FRAME-OR-WINDOW is the parent window of the selected window."
  (interactive "P")
  (let ((window (cond
		 (subtree
		  (window-parent))
		 ((windowp frame-or-window)
		  frame-or-window)
		 (t
		  (window-main-window frame-or-window)))))
    (window--transpose window '(right . above) nil)))

(defun rotate-windows-clockwise (&optional subtree frame-or-window)
  "Rotate windows of FRAME-OR-WINDOW clockwise by 90 degrees.
FRAME-OR-WINDOW must be a valid frame or window and defaults to the
selected frame.  If FRAME-OR-WINDOW is a frame, rotate the main window
of the frame, otherwise rotate FRAME-OR-WINDOW.  See
`rotate-windows-anticlockwise' for how to rotate windows in the opposite
direction.  If SUBTREE is non-nil, the function will act as if
FRAME-OR-WINDOW is the parent window of the selected window."
  (interactive "P")
  (let ((window (cond
		 (subtree
		  (window-parent))
		 ((windowp frame-or-window)
		  frame-or-window)
		 (t
		  (window-main-window frame-or-window)))))
    (window--transpose window '(left . below) nil)))

(defun flip-windows-horizontally (&optional subtree frame-or-window)
  "Horizontally flip windows of FRAME-OR-WINDOW.
When the windows are flipped horzontally, the window layout is made to
it's reflection from the side edge.  FRAME-OR-WINDOW must be a valid
frame or window and defaults to the selected frame.  If FRAME-OR-WINDOW
is a frame, flip from the main window of the frame, otherwise flip from
FRAME-OR-WINDOW.  See `flip-windows-vertically' for how to flip windows
vertically.  If SUBTREE is non-nil, the function will act as if
FRAME-OR-WINDOW is the parent window of the selected window."
  (interactive "P")
  (let ((window (cond
		 (subtree
		  (window-parent))
		 ((windowp frame-or-window)
		  frame-or-window)
		 (t
		  (window-main-window frame-or-window)))))
    (window--transpose window '(below . left) t)))

(defun flip-windows-vertically (&optional subtree frame-or-window)
  "Horizontally flip windows of FRAME-OR-WINDOW.
When the windows are flipped vertically, the window layout is made to
it's reflection from the top edge.  FRAME-OR-WINDOW must be a valid frame
or window and defaults to the selected frame.  If FRAME-OR-WINDOW is a
frame, flip from the main window of the frame, otherwise flip from
FRAME-OR-WINDOW.  See `flip-windows-horizontally' for how to flip
windows horizontally."
  (interactive "P")
  (let ((window (cond
		 (subtree
		  (window-parent))
		 ((windowp frame-or-window)
		  frame-or-window)
		 (t
		  (window-main-window frame-or-window)))))
    (window--transpose window '(above . right) t)))

(defun transpose-windows (&optional subtree frame-or-window)
  "Transpose windows of FRAME-OR-WINDOW.
Rearrange windows such that where a horizontal split was used a vertical
one is used instead, and vice versa.  FRAME-OR-WINDOW must be a valid
frame or window and defaults to the selected frame.  If FRAME-OR-WINDOW
is a frame, transpose the main window of the frame, otherwise transpose
FRAME-OR-WINDOW.  If SUBTREE is non-nil, the function will act as if
FRAME-OR-WINDOW is the parent window of the selected window."
  (interactive "P")
  (let ((window (cond
		 (subtree
		  (window-parent))
		 ((windowp frame-or-window)
		  frame-or-window)
		 (t
		  (window-main-window frame-or-window)))))
    (window--transpose window '(right . below) nil)))

(defun deepmap(fun ls)
  (if (consp ls)
      (cons
       (if (consp (car ls))
	  (deepmap fun (car ls))
	 (funcall fun (car ls)))
       (deepmap fun (cdr ls)))
    (funcall fun ls)))

(defun cycle-windows (&optional subtree frame-or-window)
  "This cycles windows basiclly so that for all windows, each window
becomes it's `other-window'.
+-----+     +-----+
|  A  |	    |  C  |
|-----| ->  |-----|
|B | C|	    | A| B|
+-----+	    +-----+
"
  (interactive "P")
  (let ((window (cond
		 (subtree
		  (window-parent))
		 ((windowp frame-or-window)
		  frame-or-window)
		 (t
		  (window-main-window frame-or-window)))))
    (if (or (not window)
	    (window-live-p window))
	(message "No windows to transpose")
      (let* ((frame (window-frame window))
	     (selwin (frame-selected-window window))
	     (win-tree (car (window-tree-pixel-sizes window)))
	     (winls (seq-filter 'window-live-p (flatten-list win-tree)))
	     (rotated-ls (append (last winls) winls))
	     (fwin (car (last winls)))
	     (new-win-tree (deepmap
			    (lambda (x)
			      (if (window-live-p x)
				  (pop rotated-ls)
				x))
			    win-tree)))
	(if (or (seq-some 'window-atom-root winls)
		(seq-some 'window-fixed-size-p winls))
	    (message "This does not work with fixed size or atom windows.")
	    (progn
	      ;; All child windows need to be recursively deleted.
	      (delete-other-windows-internal fwin window)
	      ;; (delete-dups atom-windows)
	      (window--transpose-1 new-win-tree fwin '(below . right) t nil)
	      (set-frame-selected-window frame selwin)
	      (other-window -1)
	      (while (not (memq (selected-window) winls))
		(other-window -1))))))))

(defun window--transpose (window conf do-not-convert-size)
  "Rearrange windows of WINDOW recursively.
CONF should be a cons cell: (HORIZONTAL-SPLIT . VERTICAL-SPLIT) where
HORIZONTAL-SPLIT will be used as the third argument of `split-window'
when splitting a window that was previously horizontally split, and
VERTICAL-SPLIT as third argument of `split-window' for a window that was
previously vertically split.  If DO-NOT-CONVERT-SIZE non-nil, the size
argument of the window-split is converted from vertical to horizontal or
vice versa, with the same proportion of the total split."
  (if (or (not window)
	  (window-live-p window))
      (message "No windows to transpose")
    (let* ((frame (window-frame window))
	   (fwin window)
	   (selwin (frame-selected-window window))
	   (win-tree (car (window-tree-pixel-sizes window)))
	   (win-list (seq-filter 'window-live-p (flatten-list win-tree)))
	   (atom-windows
	    (remq nil (mapcar 'window-atom-root
			      win-list))))
      (if (and (not (eq (car atom-windows) window))
	   (or do-not-convert-size
	      (and (not atom-windows)
		   (not (seq-some 'window-fixed-size-p win-list)))))
	  (progn
	    (when atom-windows
	      (delete-dups atom-windows))
	    (while (not (window-live-p fwin))
	      (setq fwin (window-child fwin)))
	    ;; All child windows need to be recursively deleted.
	    (delete-other-windows-internal fwin window)
	    (window--transpose-1 win-tree fwin conf do-not-convert-size atom-windows)
	    ;; Go back to previously selected window.
	    (set-frame-selected-window frame selwin)
	    (mapc 'window-make-atom atom-windows))
	(message "This does not work with fixed size or atom windows.")))))

(defun window--transpose-1 (subtree cwin conf do-not-convert-size atom-windows)
  "Subroutine of `window--transpose'.
SUBTREE must be in the format of the result of
`window-tree-pixel-sizes'.  CWIN is the current window through which the
window splits are made.  The CONF and DO-NOT-CONVERT-SIZE arguments are
the same as the ones in `window--transpose'."
  ;; `ilen' is the max size a window could be of given the split type.
  ;; `flen' is max size the window could be converted to the opposite
  ;; of the given split type.
  (let ((n-set t)
	(flen (if (funcall (if do-not-convert-size 'not 'identity)
			   (caar subtree))
		  (float (window-pixel-width cwin))
		(float (window-pixel-height cwin)))))
    (mapc
     (pcase-lambda (`(,win . ,size))
       (prog1
	   (let ((split-size (- (round (* flen size))))
		 (split-type
		  (funcall (if (caar subtree) 'car 'cdr) conf)))
	     (if (listp win)
		 ;; `win' is a window subtree.
		 (let* ((d-win win)
			(is-atom
			 (memq (cdar d-win) atom-windows))
			(_ (while (listp (caaddr d-win))
			     (setq d-win (caddr d-win)))))
		   (window--transpose-1
		    win
		    (let ((window-combination-limit n-set)
			  (new-win (split-window
				    cwin
				    split-size
				    split-type
				    t
				    (if window-combination-limit
					(cons (caaddr d-win) (cdar subtree))
				      (caaddr d-win)))))
		      (when window-combination-limit
			(set-window-combination-limit (cdar subtree) nil))
		      new-win)
		    (if is-atom
			'(nil . t)
		      conf)
		    do-not-convert-size
		    atom-windows))
	       ;; `win' is a window.
	       (split-window cwin split-size
			     split-type t
			     (if n-set
				 (cons win (cdar subtree))
			       win))))
	 (setq n-set nil)))
     (mapcar
      (lambda (e)
	(let ((window? (if (windowp (car e)) (car e) e)))
	  (cons window?
		;; The respective size of the window.
		(if (caar subtree)
		    (car (cadr e))
		  (cdr (cadr e))))))
      ;; By using cdddr, we ignore window split type, sizes and the
      ;; first window (it's implicitly created).
      (nreverse (cdddr subtree))))
    ;; (caaddr subtree) is the first window.
    (unless (windowp (caaddr subtree))
      (let ((is-atom (memq (cdar (caddr subtree)) atom-windows)))
	(window--transpose-1 (caddr subtree) cwin (if is-atom '(nil . t) conf)
			     do-not-convert-size atom-windows)))))

(define-key global-map (kbd "C-x w t") 'transpose-windows)
(define-key global-map (kbd "C-x w c") 'cycle-windows)
(define-key global-map (kbd "C-x w f") 'flip-windows-vertically)


